/*
 * MonitorV2.c
 *
 * Created: 01/08/2015 12:26:40
 *  Author: mike, shalewyn.com
 */ 

// Requirements: Disable the JTAGEN and CKDIV8 fuses in the uP

// indicate the speed setting of the uP, 8MHz which is why CKDIV8 needs to be disabled
#define F_CPU 8000000UL

// various include files that we need
#include <avr/io.h>
#include <util/delay.h>
#include <string.h>
#include <stdio.h>

// definitions used to select buffers/data source
#define DATABUS 0
#define ADDRBUS_LSB 1
#define ADDRBUS_MSB 2
#define NOBUS 3
#define SIGNALS 4

#define LCDcmd_displayON 0b00001100
#define LCDcmd_displayOFF 0b00001000
#define LCDcmd_clear 0b00000001
#define LCDcmd_home 0b00000010

/*
 Port setup
 ----------

 Remember to untick/disable JTAGEN fuse as we need to use the pins
 as assigned to JTAG and we're using SPI to program the AVR
 
 PA0 to PA7 = 6502 bus (input)

 PC0 = 6502 address bus LSB buffer /enable (output)
 PC1 = 6502 address bus MSB buffer /enable (output)
 PC2 = 6502 data bus buffer /enable (output)
 PC3 = 6502 signals buffer /enable (output)
 PC5 = LCD E (output)
 PC6 = LCD RW (input/output)
 PC7 = LCD RS (output)

 PD0 to PD7 = LCD DB0 to DB7 (output)

 PB5 = MOSI (input/output) - RESERVED
 PB6 = MISO (input/output) - RESERVED
 PB7 = SCK (input/output) - RESERVED
 
 */


// some macros to handle bit-wise stuff. kindly provided by AVRfreaks.net
// (http://www.avrfreaks.net/index.php?name=PNphpBB2&file=viewtopic&p=40348#40348)
#define bit_get(p,m) ((p) & (m))
#define bit_set(p,m) ((p) |= (m))
#define bit_clear(p,m) ((p) &= ~(m))
#define bit_flip(p,m) ((p) ^= (m))
#define bit_write(c,p,m) (c ? bit_set(p,m) : bit_clear(p,m))
#define BIT(x) (0x01 << (x))
#define LONGBIT(x) ((unsigned lONg)0x00000001 << (x))

void BufferSelect(int d);


// set up the various ports that we need for input/output
void PortSetup(void)
{
	DDRD = 0b11111111;		// Set portD as output
	PORTD = 0x0;			// Set portD output to 0

	DDRC = 0b11111111;		// Set portC as output
	PORTC = 0b11111111;		// Set the outputting data on portC
	BufferSelect(NOBUS);
	
	DDRA = 0b00000000;		// set portA as input
	PORTA = 0xFF;			// enable pull up resistors for port A
	
	return;
}

/*
	LCD command time requirements

	Clear Display				15.2ms
	Return Home					15.2ms
	Entry Mode Set				40us
	Display on/off				40us
	Cursor/display shift		40us
	Function set				40us
	Set GC RAM Address			40us
	Set DD RAM Address			40us
	Read Busy Flag & Address	40us
	Write data to GC or DD RAM	40us
	Read data from GC or DD RAM	40us
*/

// LCD routine for sending a command byte
void LCDsendcmd(unsigned char d)
{
	bit_clear(PORTC,BIT(7));	// command mode
	PORTD = d;
	bit_set(PORTC,BIT(5));		// send E=1
	bit_clear(PORTC,BIT(5));	// send E=0
	switch (d)
	{
		case LCDcmd_clear:
		case LCDcmd_home:
			_delay_ms(16);
			break;		
		default:	
			_delay_us(45);
			break;
	}
	bit_set(PORTC,BIT(5));		// send E=1
}

// LCD routine for sending a character/data byte
void LCDsenddata(unsigned char d)
{
	bit_set(PORTC,BIT(7));		// data mode
	PORTD = d;
	bit_set(PORTC,BIT(5));		// send E=1
	bit_clear(PORTC,BIT(5));	// send E=0
	_delay_ms(3);
	bit_set(PORTC,BIT(5));		// send E=1
	_delay_ms(3);
}

// LCD routine for sending a string of characters using LCDsenddata
void LCDsendstr(char *s)
{
	int i;
	for (i=0; i<strlen(s); i++)
	{
		LCDsenddata(s[i]);
	}	
}

// LCD Initialisation routine
void LCDinit(void)
{
	// RS = PC7, E=PC5, PC6=RW
	// RS - 0=command, 1=data
	// RW - 0
	// E - transit high to low to action command/data
	
	bit_set(PORTC, BIT(5));		// don't action yet
	bit_clear(PORTC, BIT(6));	// write mode
	bit_clear(PORTC, BIT(7));	// command mode
	
	_delay_ms(15);	
	PORTD = 0b00110000;			// Send 0011 0000 three times
	bit_clear(PORTC, BIT(5));	// send E=0
	_delay_ms(5);
	bit_set(PORTC, BIT(5));		// send E=1

	bit_clear(PORTC, BIT(5));	// send E=0
	_delay_us(120);
	bit_set(PORTC, BIT(5));		// send E=1

	bit_clear(PORTC,BIT(5));	// send E=0
	_delay_ms(3);
	bit_set(PORTC,BIT(5));		// send E=1

	LCDsendcmd(0b00111000);		// Function Set to set 8bit mode
	LCDsendcmd(0b00111000);		// Function Set to set no. of lines and font
	LCDsendcmd(0b00001000);		// display off
	LCDsendcmd(0b00000001);		// clear display
	LCDsendcmd(0b00000110);		// entry mode

	return;	
}

// LCD cursor movement routine
void LCDgotoXY(int x, int y)
{
	unsigned char lcd_offset = 0;

	// correct any value outside the normal ranges
	if (x < 1) x = 1;
	if (y < 1) y = 1;
	if (x > 20) x = 20;
	if (y > 4) y = 4;

	// get the LCD RAM offset for the specified row/line
	switch (y)
	{
		case 1:
		lcd_offset = 0x00;
		break;
		case 2:
		lcd_offset = 0x40;
		break;
		case 3:
		lcd_offset = 0x14;
		break;
		case 4:
		lcd_offset = 0x54;
		break;
	}
	lcd_offset = lcd_offset + x - 1;	// add in the x coordinate minus 1 to set the column
	bit_set(lcd_offset, BIT(7));		// "Set DD RAM Address" command used by setting bit 7
	LCDsendcmd(lcd_offset);				// send command
	
}

// Routine for selecting the required input buffer/data source
void BufferSelect(int bufferID)
{
	/*
		PC0 = 6502 address bus LSB buffer /enable (output)
		PC1 = 6502 address bus MSB buffer /enable (output)
		PC2 = 6502 data bus buffer /enable (output)
		PC3 = 6502 signals buffer /enable (output)
	*/

	// enable buffer bufferID, but ensure that all other buffers are
	// disabled to avoid bus contention. this is why the enabling
	// is done as the last action.
	
	switch(bufferID)
	{
		case DATABUS:
			bit_set(PORTC,BIT(0));
			bit_set(PORTC,BIT(1));
			bit_set(PORTC,BIT(3));
			bit_clear(PORTC,BIT(2));
			break;
		case ADDRBUS_LSB:
			bit_set(PORTC,BIT(1));
			bit_set(PORTC,BIT(2));
			bit_set(PORTC,BIT(3));
			bit_clear(PORTC,BIT(0));
			break;
		case ADDRBUS_MSB:
			bit_set(PORTC,BIT(0));
			bit_set(PORTC,BIT(2));
			bit_set(PORTC,BIT(3));
			bit_clear(PORTC,BIT(1));
			break;
		case NOBUS:
			bit_set(PORTC,BIT(0));
			bit_set(PORTC,BIT(1));
			bit_set(PORTC,BIT(2));
			bit_set(PORTC,BIT(3));
			break;
		case SIGNALS:
			bit_set(PORTC,BIT(0));
			bit_set(PORTC,BIT(1));
			bit_set(PORTC,BIT(2));
			bit_clear(PORTC,BIT(3));
			break;
	}
}


int main(void)
{
	// define the variables that are needed		
	unsigned char addrLSBbus;
	unsigned char addrMSBbus;
	unsigned char databus;
	int init_display = 1;
	
	char sd[21] = {0};
	char sd2[21] = {0};
	char sd3[6] = {0};
	char sd4[6] = {0};
	char sd5[6] = {0};
	char sd6[6] = {0};
	char sd7[6] = {0};
	char sd8[6] = {0};
	char sd9[6] = {0};
		
	unsigned char signals_byte = 0;
	char bitarray[10] = {0};
	int loop = 0;

	// set things up	
	PortSetup();
	LCDinit();

	// display the startup banner. keep in mind that the display is deactivated at this time
	LCDgotoXY(1,1); LCDsendstr("********************");
	LCDgotoXY(1,2);	LCDsendstr("** BUS MONITOR V2 **");
	LCDgotoXY(1,3);	LCDsendstr("**  shalewyn.com  **");
	LCDgotoXY(1,4);	LCDsendstr("********************");

	// now activate the display	
	LCDsendcmd(0b00001100);		// display on
	_delay_ms(3000);			// wait 3 seconds
	LCDsendcmd(0b00001000);		// display off
	LCDsendcmd(0b00000001);		// clear display
	
	// display the on screen labels
	LCDgotoXY(1,1);
	LCDsendstr("Address bus $");
	LCDgotoXY(1,2);
	LCDsendstr("   Data bus $");
		
	while(1)
    {
		// gather the data that we need, one set of 8 bits at a time.
		// we do this be selecting/enabling each buffer IC in turn
		// and reading PORTA
		
		BufferSelect(ADDRBUS_LSB);
		addrLSBbus = PINA;
		
		BufferSelect(ADDRBUS_MSB);
		addrMSBbus = PINA;
		
		BufferSelect(DATABUS);
		databus = PINA;
		
		BufferSelect(SIGNALS);
		signals_byte = PINA;	
		BufferSelect(NOBUS);	
				
		// grab PortA values and separate it in to a bit array
		for (loop=0; loop<8; loop++)
		{
			bitarray[loop] = signals_byte & 1;
			signals_byte >>= 1;
			//if (bitarray[loop]==0) strcat(ss, "0 "); else strcat(ss, "1 ");
		}	

		// now that we have the data we cam build strings which incorporate it 
		// so that we can display it
		sprintf(sd, "%2x%2x", addrMSBbus, addrLSBbus);
		sprintf(sd2, "%2x", databus);

		// check to see if we need to inject any leading zero's into the address bus LSB & MSB
		// and the data bus. Otherwise we'll end up with a single digit instead of 2 digits if
		// the value is 15 (0xF) or below.
		if (addrMSBbus <= 15) sd[0] = 48;
		if (addrLSBbus <= 15) sd[2] = 48;
		if (databus <= 15) sd2[0] = 48;

		// Decide which signals need to be displayed
		if (bitarray[0]==0) sprintf(sd3, "ROM ");  else sprintf(sd3, "    ");
		if (bitarray[1]==0) sprintf(sd4, "VIA1");  else sprintf(sd4, "    ");
		if (bitarray[2]==0) sprintf(sd5, "RAM ");  else sprintf(sd5, "    ");
		if (bitarray[3]==0) sprintf(sd6, "VIA2");  else sprintf(sd6, "    ");
		if (bitarray[4]==0) sprintf(sd7, "ACIA"); else sprintf(sd7, "    ");
		if (bitarray[5]==0) sprintf(sd8, "BNK0"); else sprintf(sd8, "    ");
		if (bitarray[6]==0) sprintf(sd9, "BNK1"); else sprintf(sd9, "    ");

		// display the Address bus and Data bus values
		LCDgotoXY(14,1);	LCDsendstr(sd);
		LCDgotoXY(14,2);	LCDsendstr(sd2);
		
		// display the signals 1st line: ROM, RAM, VIA1, VIA2
		LCDgotoXY(1,3);		LCDsendstr(sd3);
		LCDgotoXY(6,3);		LCDsendstr(sd5);
		LCDgotoXY(11,3);	LCDsendstr(sd4);
		LCDgotoXY(16,3);	LCDsendstr(sd6);

		// display the signals 2nd line: BNK0, BNK1, ACIA
		LCDgotoXY(1,4);		LCDsendstr(sd8);
		LCDgotoXY(6,4);		LCDsendstr(sd9);
		LCDgotoXY(16,4);	LCDsendstr(sd7);

		// the first time around, the above is sent to the LCD while the
		// display is off (for speed). Now the display needs to be turned
		// back on (only runs once)
		if (init_display)
		{
			LCDsendcmd(0b00001100);						// display on
			init_display = 1;	
		}
	}
}